/******************************************************************************/
/*                                                                            */
/*                Advanced Navigation Packet Protocol Library                 */
/*                 Java Language Spatial FOG SDK, Version 6.0                 */
/*                    Copyright 2022, Advanced Navigation                     */
/*                                                                            */
/******************************************************************************/
/*
 * Copyright (C) 2022 Advanced Navigation
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

package com.advancednavigation.anPackets;

import java.io.UnsupportedEncodingException;
import java.util.Date;
import java.util.Locale;

/* GNSS Receiver Information */
public class ANPacket69
{
	public static final int GNSS_MANUFACTURER_UNKNOWN = 0;
	public static final int GNSS_MANUFACTURER_TRIMBLE = 1;
	public static final int GNSS_MANUFACTURER_UBLOX = 2;
	public static final int GNSS_MANUFACTURER_ADNAV = 3;

	public static final int GNSS_RECEIVER_MODEL_UNKNOWN = 0;
	public static final int GNSS_RECEIVER_MODEL_TRIMBLE_BD920 = 1;
	public static final int GNSS_RECEIVER_MODEL_TRIMBLE_BD930 = 2;
	public static final int GNSS_RECEIVER_MODEL_TRIMBLE_BD982 = 3;
	public static final int GNSS_RECEIVER_MODEL_TRIMBLE_MBONE = 4;
	public static final int GNSS_RECEIVER_MODEL_TRIMBLE_MBTWO = 5;
	public static final int GNSS_RECEIVER_MODEL_TRIMBLE_BD940 = 6;
	public static final int GNSS_RECEIVER_MODEL_TRIMBLE_BD992 = 7;
	public static final int GNSS_RECEIVER_MODEL_UBLOX_M8  = 1;
	public static final int GNSS_RECEIVER_MODEL_UBLOX_M8T = 2;
	public static final int GNSS_RECEIVER_MODEL_UBLOX_M8P = 3;
	public static final int GNSS_RECEIVER_MODEL_UBLOX_F9P = 4;
	public static final int GNSS_RECEIVER_MODEL_ADNAV_ARIES = 1;

	public static final int OMNISTAR_ENGINE_MODE_INACTIVE = 0;
	public static final int OMNISTAR_ENGINE_MODE_HP = 1;
	public static final int OMNISTAR_ENGINE_MODE_XP = 2;
	public static final int OMNISTAR_ENGINE_MODE_G2 = 3;
	public static final int OMNISTAR_ENGINE_MODE_HP_G2 = 4;
	public static final int OMNISTAR_ENGINE_MODE_HP_XP = 5;

	public static final int RTK_ACCURACY_UNKNOWN = 0;
	public static final int RTK_ACCURACY_300_300 = 1;
	public static final int RTK_ACCURACY_70_70 = 2;
	public static final int RTK_ACCURACY_70_20 = 3;
	public static final int RTK_ACCURACY_8_70 = 4;
	public static final int RTK_ACCURACY_8_20 = 5;

	public static final int LENGTH = 68;
	public static final int LENGTH_LEGACY = 48;

	public int gnssManufacturerID;
	public int gnssReceiverModel;
	public String gnssSerialNumber;
	public float firmwareVersion;
	public float hardwareVersion;
	public String softwareLicense;
	public String partNumber;
	public long omnistarSerialNumber;
	public Date omnistarSubscriptionStart;
	public Date omnistarSubscriptionExpiry;
	public int omnistarEngineMode;
	public int rtkAccuracy;

	public ANPacket69(ANPacket packet)
	{
		if (packet.length == LENGTH)
		{
			// standard definition
			gnssManufacturerID = packet.data[0] & 0xFF;
			gnssReceiverModel = packet.data[1] & 0xFF;
			try
			{
				gnssSerialNumber = new String(packet.data, 2, 24, "US-ASCII");
			}
			catch (UnsupportedEncodingException e)
			{
				gnssSerialNumber = "";
			}
			firmwareVersion = TypeConversion.bytesToUInt32(packet.data, 26) / 1000.0f;
			hardwareVersion = TypeConversion.bytesToUInt32(packet.data, 30) / 1000.0f;

			long a = TypeConversion.bytesToUInt32(packet.data, 34);
			long b = TypeConversion.bytesToUInt32(packet.data, 38);
			long c = TypeConversion.bytesToUInt32(packet.data, 42);
			softwareLicense = String.format(Locale.ENGLISH, "%08X%08X%08X", a, b, c);
			omnistarSerialNumber = TypeConversion.bytesToUInt32(packet.data, 46);
			omnistarSubscriptionStart = new Date(TypeConversion.bytesToUInt32(packet.data, 50) * 1000);
			omnistarSubscriptionExpiry = new Date(TypeConversion.bytesToUInt32(packet.data, 54) * 1000);
			omnistarEngineMode = packet.data[58] & 0xFF;
			rtkAccuracy = packet.data[59] & 0xFF;
		}
		else if (packet.length == LENGTH_LEGACY)
		{
			// legacy definition
			gnssManufacturerID = packet.data[0] & 0xFF;
			gnssReceiverModel = packet.data[1] & 0xFF;
			try
			{
				gnssSerialNumber = new String(packet.data, 2, 10, "US-ASCII");
			}
			catch (UnsupportedEncodingException e)
			{
				gnssSerialNumber = "";
			}
			firmwareVersion = TypeConversion.bytesToUInt32(packet.data, 12) / 1000.0f;
			hardwareVersion = 0.0f;

			long a = TypeConversion.bytesToUInt32(packet.data, 16);
			long b = TypeConversion.bytesToUInt32(packet.data, 20);
			long c = TypeConversion.bytesToUInt32(packet.data, 24);
			softwareLicense = String.format(Locale.ENGLISH, "%08X%08X%08X", a, b, c);
			omnistarSerialNumber = TypeConversion.bytesToUInt32(packet.data, 28);
			omnistarSubscriptionStart = new Date(TypeConversion.bytesToUInt32(packet.data, 32) * 1000);
			omnistarSubscriptionExpiry = new Date(TypeConversion.bytesToUInt32(packet.data, 36) * 1000);
			omnistarEngineMode = packet.data[40] & 0xFF;
			rtkAccuracy = packet.data[41] & 0xFF;
		}
	}

	public String receiverModelDescription()
	{
		if (gnssManufacturerID == GNSS_MANUFACTURER_TRIMBLE)
		{
			switch (gnssReceiverModel)
			{
				case GNSS_RECEIVER_MODEL_TRIMBLE_BD920:
					return "Trimble BD920";
				case GNSS_RECEIVER_MODEL_TRIMBLE_BD930:
					return "Trimble BD930";
				case GNSS_RECEIVER_MODEL_TRIMBLE_BD982:
					return "Trimble BD982";
				case GNSS_RECEIVER_MODEL_TRIMBLE_MBONE:
					return "Trimble MB-One";
				case GNSS_RECEIVER_MODEL_TRIMBLE_MBTWO:
					return "Trimble MB-Two";
				case GNSS_RECEIVER_MODEL_TRIMBLE_BD940:
					return "Trimble BD940";
				case GNSS_RECEIVER_MODEL_TRIMBLE_BD992:
					return "Trimble BD992";
				default:
					return "Trimble Unknown";
			}
		}
		else if (gnssManufacturerID == GNSS_MANUFACTURER_UBLOX)
		{
			switch (gnssReceiverModel)
			{
				case GNSS_RECEIVER_MODEL_UBLOX_M8:
					return "Ublox ZED-M8";
				case GNSS_RECEIVER_MODEL_UBLOX_M8T:
					return "Ublox ZED-M8T";
				case GNSS_RECEIVER_MODEL_UBLOX_M8P:
					return "Ublox ZED-M8P";
				case GNSS_RECEIVER_MODEL_UBLOX_F9P:
					return "Ublox ZED-F9P";
				default:
					return "Ublox Unknown";
			}
		}
		else if (gnssManufacturerID == GNSS_MANUFACTURER_ADNAV)
		{
			switch (gnssReceiverModel)
			{
				case GNSS_RECEIVER_MODEL_ADNAV_ARIES:
					return "Advanced Navigation Aries";
				default:
					return "Advanced Navigation Unknown";
			}
		}
		else
		{
			return "Unknown";
		}
	}

	public String receiverManufDescription()
	{
		switch (gnssManufacturerID)
		{
			case GNSS_MANUFACTURER_TRIMBLE:
				return "Trimble";
			case GNSS_MANUFACTURER_UBLOX:
				return "UBlox";
			case GNSS_MANUFACTURER_ADNAV:
				return "Advanced Navigation";
			default:
				return "Unknown";
		}
	}

	public String omnistarEngineModeDescription()
	{
		switch (omnistarEngineMode)
		{
			case OMNISTAR_ENGINE_MODE_INACTIVE:
				return "Not Active";
			case OMNISTAR_ENGINE_MODE_HP:
				return "HP";
			case OMNISTAR_ENGINE_MODE_XP:
				return "XP";
			case OMNISTAR_ENGINE_MODE_G2:
				return "G2";
			case OMNISTAR_ENGINE_MODE_HP_G2:
				return "HP+G2";
			case OMNISTAR_ENGINE_MODE_HP_XP:
				return "HP+XP";
			default:
				return "Unknown Mode " + omnistarEngineMode;
		}
	}

	public String rtkDitheringDescription()
	{
		switch (rtkAccuracy)
		{
			case RTK_ACCURACY_UNKNOWN:
				return "None";
			case RTK_ACCURACY_300_300:
				return "0.3m Horizontal, 0.3m Vertical";
			case RTK_ACCURACY_70_70:
				return "0.1m Horizontal, 0.1m Vertical";
			case RTK_ACCURACY_70_20:
				return "0.1m Horizontal, 0.02m Vertical";
			case RTK_ACCURACY_8_70:
				return "0.008m Horizontal, 0.1m Vertical";
			case RTK_ACCURACY_8_20:
				return "0.008m Horizontal, 0.02m Vertical";
			default:
				return "Unknown";
		}
	}
}
